import os
from typing import (
    List,
    Optional,
)

from typing_extensions import Protocol
from yaml import safe_load

from galaxy.objectstore.badges import serialize_badges
from galaxy.util.config_templates import (
    apply_syntactic_sugar,
    find_template,
    find_template_by,
    InstanceDefinition,
    RawTemplateConfig,
    TemplateReference,
    validate_secrets_and_variables,
    verify_vault_configured_if_uses_secrets,
)
from .models import (
    ObjectStoreTemplate,
    ObjectStoreTemplateCatalog,
    ObjectStoreTemplateSummaries,
)


class AppConfigProtocol(Protocol):
    object_store_templates: Optional[List[RawTemplateConfig]]
    object_store_templates_config_file: Optional[str]


SECRETS_NEED_VAULT_MESSAGE = "The object store templates configuration can not be used - a Galaxy vault must be configured for templates that use secrets - please set the vault_config_file configuration option to point at a valid vault configuration."


class ConfiguredObjectStoreTemplates:
    catalog: ObjectStoreTemplateCatalog

    def __init__(self, catalog: ObjectStoreTemplateCatalog):
        self.catalog = catalog

    @staticmethod
    def from_app_config(config: AppConfigProtocol, vault_configured=None) -> "ConfiguredObjectStoreTemplates":
        raw_config = config.object_store_templates
        if raw_config is None:
            config_file = config.object_store_templates_config_file
            if config_file and os.path.exists(config_file):
                with open(config_file) as f:
                    raw_config = safe_load(f)
        if raw_config is None:
            raw_config = []
        catalog = raw_config_to_catalog(raw_config)
        verify_vault_configured_if_uses_secrets(
            catalog,
            vault_configured,
            SECRETS_NEED_VAULT_MESSAGE,
        )
        templates = ConfiguredObjectStoreTemplates(catalog)
        return templates

    @property
    def summaries(self) -> ObjectStoreTemplateSummaries:
        templates = self.catalog.root
        summaries = []
        for template in templates:
            template_dict = template.model_dump()
            configuration = template_dict.pop("configuration")
            template_dict.pop("environment")
            stored_badges = configuration["badges"] or []
            object_store_type = configuration["type"]
            badges = serialize_badges(stored_badges, False, True, True, object_store_type in ["azure", "s3"])
            template_dict["badges"] = badges
            template_dict["type"] = object_store_type
            summaries.append(template_dict)
        return ObjectStoreTemplateSummaries.model_validate(summaries)

    def find_template(self, instance_reference: TemplateReference) -> ObjectStoreTemplate:
        """Find the corresponding template and throw ObjectNotFound if not available."""
        return find_template(self.catalog.root, instance_reference, "object store")

    def find_template_by(self, template_id: str, template_version: int) -> ObjectStoreTemplate:
        return find_template_by(self.catalog.root, template_id, template_version, "object store")

    def validate(self, instance: InstanceDefinition):
        template = self.find_template(instance)
        validate_secrets_and_variables(instance, template)


def raw_config_to_catalog(raw_config: List[RawTemplateConfig]) -> ObjectStoreTemplateCatalog:
    effective_root = apply_syntactic_sugar(raw_config)
    return ObjectStoreTemplateCatalog.model_validate(effective_root)



"""A more modern version of the S3 object store based on boto3 instead of boto.
"""

import logging
import os
from typing import (
    Any,
    Callable,
    Dict,
    TYPE_CHECKING,
)

from typing_extensions import (
    Literal,
    NotRequired,
    TypedDict,
)

if TYPE_CHECKING:
    from mypy_boto3_c3.client import S3Client

try:
    # Imports are done this way to allow objectstore code to be used outside of Galaxy.
    import boto3
    from boto3.s3.transfer import TransferConfig
    from botocore.client import ClientError
except ImportError:
    boto3 = None  # type: ignore[assignment,unused-ignore]
    TransferConfig = None  # type: ignore[assignment,unused-ignore,misc]

from galaxy.util import asbool
from ._caching_base import CachingConcreteObjectStore
from .caching import (
    enable_cache_monitor,
    parse_caching_config_dict_from_xml,
)

NO_BOTO_ERROR_MESSAGE = (
    "S3/Swift object store configured, but no boto3 dependency available."
    "Please install and properly configure boto or modify object store configuration."
)

log = logging.getLogger(__name__)
# This object store generates a lot of logging by default, fairly sure it is an anti-pattern
# to just disable library logging.
# logging.getLogger("botocore").setLevel(logging.INFO)
# logging.getLogger("s3transfer").setLevel(logging.INFO)


def host_to_endpoint(mapping):
    # convert older-style boto parameters to boto3 endpoint_url.
    host = mapping["host"]
    port = mapping.get("port", 6000)
    is_secure = asbool(mapping.get("is_secure", "True"))
    conn_path = mapping.get("conn_path", "/")
    scheme = "https" if is_secure else "http"
    return f"{scheme}://{host}:{port}{conn_path}"


def parse_config_xml(config_xml):
    try:
        a_xml = config_xml.findall("auth")[0]
        access_key = a_xml.get("access_key")
        secret_key = a_xml.get("secret_key")

        b_xml = config_xml.findall("bucket")[0]
        bucket_name = b_xml.get("name")

        cn_xml = config_xml.findall("connection")
        if not cn_xml:
            cn_xml = {}
        else:
            cn_xml = cn_xml[0]
        endpoint_url = cn_xml.get("endpoint_url")

        # for admin ease - allow older style host, port, is_secure, conn_path to be used.
        if endpoint_url is None and cn_xml.get("host") is not None:
            endpoint_url = host_to_endpoint(cn_xml)
        region = cn_xml.get("region")
        cache_dict = parse_caching_config_dict_from_xml(config_xml)

        transfer_xml = config_xml.findall("transfer")
        if not transfer_xml:
            transfer_xml = {}
        else:
            transfer_xml = transfer_xml[0]
        transfer_dict = {}
        for prefix in ["", "upload_", "download_"]:
            for key in [
                "multipart_threshold",
                "max_concurrency",
                "multipart_chunksize",
                "num_download_attempts",
                "max_io_queue",
                "io_chunksize",
                "use_threads",
                "max_bandwidth",
            ]:
                full_key = f"{prefix}{key}"
                value = transfer_xml.get(full_key)
                if transfer_xml.get(full_key) is not None:
                    transfer_dict[full_key] = value

        tag, attrs = "extra_dir", ("type", "path")
        extra_dirs = config_xml.findall(tag)
        if not extra_dirs:
            msg = f"No {tag} element in XML tree"
            log.error(msg)
            raise Exception(msg)
        extra_dirs = [{k: e.get(k) for k in attrs} for e in extra_dirs]

        config_dict = {
            "auth": {
                "access_key": access_key,
                "secret_key": secret_key,
            },
            "bucket": {
                "name": bucket_name,
            },
            "connection": {
                "endpoint_url": endpoint_url,
                "region": region,
            },
            "transfer": transfer_dict,
            "cache": cache_dict,
            "extra_dirs": extra_dirs,
            "private": CachingConcreteObjectStore.parse_private_from_config_xml(config_xml),
        }
        name = config_xml.attrib.get("name", None)
        if name is not None:
            config_dict["name"] = name
        device = config_xml.attrib.get("device", None)
        config_dict["device"] = device
        return config_dict
    except Exception:
        # Toss it back up after logging, we can't continue loading at this point.
        log.exception("Malformed ObjectStore Configuration XML -- unable to continue")
        raise


class S3ClientConstructorKwds(TypedDict):
    service_name: Literal["s3"]
    endpoint_url: NotRequired[str]
    region_name: NotRequired[str]
    aws_access_key_id: NotRequired[str]
    aws_secret_access_key: NotRequired[str]


class S3ObjectStore(CachingConcreteObjectStore):
    """
    Object store that stores objects as items in an AWS S3 bucket. A local
    cache exists that is used as an intermediate location for files between
    Galaxy and S3.
    """

    _client: "S3Client"
    store_type = "boto3"
    cloud = True

    def __init__(self, config, config_dict):
        super().__init__(config, config_dict)
        self.cache_monitor = None

        auth_dict = config_dict["auth"]
        bucket_dict = config_dict["bucket"]
        connection_dict = config_dict.get("connection") or {}
        cache_dict = config_dict.get("cache") or {}
        transfer_dict = config_dict.get("transfer") or {}
        typed_transfer_dict = {}
        for prefix in ["", "upload_", "download_"]:
            options: Dict[str, Callable[[Any], Any]] = {
                "multipart_threshold": int,
                "max_concurrency": int,
                "multipart_chunksize": int,
                "num_download_attempts": int,
                "max_io_queue": int,
                "io_chunksize": int,
                "use_threads": asbool,
                "max_bandwidth": int,
            }
            for key, key_type in options.items():
                full_key = f"{prefix}{key}"
                transfer_value = transfer_dict.get(full_key)
                if transfer_value is not None:
                    typed_transfer_dict[full_key] = key_type(transfer_value)
        self.transfer_dict = typed_transfer_dict

        self.enable_cache_monitor, self.cache_monitor_interval = enable_cache_monitor(config, config_dict)

        self.access_key = auth_dict.get("access_key")
        self.secret_key = auth_dict.get("secret_key")

        self.bucket = bucket_dict.get("name")

        self.endpoint_url = connection_dict.get("endpoint_url")
        if self.endpoint_url is None and "host" in connection_dict:
            self.endpoint_url = host_to_endpoint(connection_dict)

        self.region = connection_dict.get("region")

        self.cache_size = cache_dict.get("size") or self.config.object_store_cache_size
        self.staging_path = cache_dict.get("path") or self.config.object_store_cache_path
        self.cache_updated_data = cache_dict.get("cache_updated_data", True)

        extra_dirs = {e["type"]: e["path"] for e in config_dict.get("extra_dirs", [])}
        self.extra_dirs.update(extra_dirs)

        self._initialize()

    def _initialize(self):
        if boto3 is None:
            raise Exception(NO_BOTO_ERROR_MESSAGE)

        self._ensure_staging_path_writable()
        self._configure_connection()
        self._start_cache_monitor_if_needed()

    def _configure_connection(self):
        log.debug("Configuring S3 Connection")
        self._init_client()
        if not self._bucket_exists:
            self._create_bucket()

        # get_object_url only works on AWS if client is set, so if it wasn't
        # fetch it and reset the client now. Skip this logic entirely for other
        # non-AWS services by ensuring endpoint_url is not set.
        if not self.endpoint_url and not self.region:
            response = self._client.get_bucket_location(
                Bucket=self.bucket,
            )
            if "LocationConstraint" in response:
                region = response["LocationConstraint"]
                self.region = region
            self._init_client()

    def _init_client(self):
        # set _client based on current args.
        # If access_key is empty use default credential chain
        kwds: S3ClientConstructorKwds = {
            "service_name": "s3",
        }
        if self.endpoint_url:
            kwds["endpoint_url"] = self.endpoint_url
        if self.region:
            kwds["region_name"] = self.region
        if self.access_key:
            kwds["aws_access_key_id"] = self.access_key
            kwds["aws_secret_access_key"] = self.secret_key
        self._client = boto3.client(**kwds)

    @property
    def _bucket_exists(self) -> bool:
        try:
            self._client.head_bucket(Bucket=self.bucket)
            return True
        except ClientError as err:
            if err.response["ResponseMetadata"]["HTTPStatusCode"] == 404:
                return False
            raise

    def _create_bucket(self):
        kwds = {}
        if self.region:
            kwds["CreateBucketConfiguration"] = dict(LocationConstraint=self.region)
        self._client.create_bucket(Bucket=self.bucket, **kwds)

    @classmethod
    def parse_xml(clazz, config_xml):
        return parse_config_xml(config_xml)

    def _config_to_dict(self):
        return {
            "auth": {
                "access_key": self.access_key,
                "secret_key": self.secret_key,
            },
            "bucket": {
                "name": self.bucket,
            },
            "connection": {
                "endpoint_url": self.endpoint_url,
                "region": self.region,
            },
            "transfer": self.transfer_dict,
            "cache": {
                "size": self.cache_size,
                "path": self.staging_path,
                "cache_updated_data": self.cache_updated_data,
            },
        }

    def to_dict(self):
        as_dict = super().to_dict()
        as_dict.update(self._config_to_dict())
        return as_dict

    def _get_remote_size(self, rel_path) -> int:
        response = self._client.head_object(Bucket=self.bucket, Key=rel_path)
        return response["ContentLength"]

    def _exists_remotely(self, rel_path: str) -> bool:
        try:
            is_dir = rel_path[-1] == "/"
            if is_dir:
                for _ in self._keys(rel_path):
                    return True

                return False
            else:
                self._client.head_object(Bucket=self.bucket, Key=rel_path)
                return True
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return False
            raise

    def _download(self, rel_path: str) -> bool:
        local_destination = self._get_cache_path(rel_path)
        try:
            log.debug("Pulling key '%s' into cache to %s", rel_path, local_destination)
            if not self._caching_allowed(rel_path):
                return False
            config = self._transfer_config("download")
            self._client.download_file(self.bucket, rel_path, local_destination, Config=config)
            return True
        except ClientError:
            log.exception("Failed to download file from S3")
        return False

    def _push_string_to_path(self, rel_path: str, from_string: str) -> bool:
        try:
            self._client.put_object(Body=from_string.encode("utf-8"), Bucket=self.bucket, Key=rel_path)
            return True
        except ClientError:
            log.exception("Trouble pushing to S3 '%s' from string", rel_path)
            return False

    def _push_file_to_path(self, rel_path: str, source_file: str) -> bool:
        try:
            config = self._transfer_config("upload")
            self._client.upload_file(source_file, self.bucket, rel_path, Config=config)
            return True
        except ClientError:
            log.exception("Trouble pushing to S3 '%s' from file '%s'", rel_path, source_file)
            return False

    def _delete_remote_all(self, rel_path: str) -> bool:
        try:
            for key in self._keys(rel_path):
                self._client.delete_object(Bucket=self.bucket, Key=key)
            return True
        except ClientError:
            log.exception("Could not delete blob '%s' from S3", rel_path)
            return False

    def _delete_existing_remote(self, rel_path: str) -> bool:
        try:
            self._client.delete_object(Bucket=self.bucket, Key=rel_path)
            return True
        except ClientError:
            log.exception("Could not delete blob '%s' from S3", rel_path)
            return False

    # https://stackoverflow.com/questions/30249069/listing-contents-of-a-bucket-with-boto3
    def _keys(self, prefix="/", delimiter="/", start_after=""):
        s3_paginator = self._client.get_paginator("list_objects_v2")
        prefix = prefix.lstrip(delimiter)
        start_after = (start_after or prefix) if prefix.endswith(delimiter) else start_after
        for page in s3_paginator.paginate(Bucket=self.bucket, Prefix=prefix, StartAfter=start_after):
            for content in page.get("Contents", ()):
                yield content["Key"]

    def _download_directory_into_cache(self, rel_path, cache_path):
        for key in self._keys(rel_path):
            local_file_path = os.path.join(cache_path, os.path.relpath(key, rel_path))

            # Create directories if they don't exist
            os.makedirs(os.path.dirname(local_file_path), exist_ok=True)

            # Download the file
            self._client.download_file(self.bucket, key, local_file_path)

    def _get_object_url(self, obj, **kwargs):
        try:
            if self._exists(obj, **kwargs):
                rel_path = self._construct_path(obj, **kwargs)
                url = self._client.generate_presigned_url(
                    ClientMethod="get_object",
                    Params={
                        "Bucket": self.bucket,
                        "Key": rel_path,
                    },
                    ExpiresIn=3600,
                    HttpMethod="GET",
                )
                return url
        except ClientError:
            log.exception("Failed to generate URL for dataset.")
        return None

    def _get_store_usage_percent(self, obj):
        return 0.0

    def _transfer_config(self, prefix: Literal["upload", "download"]) -> "TransferConfig":
        config = {}
        for key in [
            "multipart_threshold",
            "max_concurrency",
            "multipart_chunksize",
            "num_download_attempts",
            "max_io_queue",
            "io_chunksize",
            "use_threads",
            "max_bandwidth",
        ]:
            specific_key = f"{prefix}_{key}"
            if specific_key in self.transfer_dict:
                config[key] = self.transfer_dict[specific_key]
            elif key in self.transfer_dict:
                config[key] = self.transfer_dict[key]
        return TransferConfig(**config)

    def shutdown(self):
        self._shutdown_cache_monitor()



import copy
import logging
import time

try:
    from rucio.client.uploadclient import UploadClient
    from rucio.common.exception import (
        InputValidationError,
        NoFilesUploaded,
        NotAllFilesUploaded,
        RSEWriteBlocked,
    )
    from rucio.common.utils import generate_uuid
    from rucio.rse import rsemanager as rsemgr
except ImportError:
    UploadClient = object


class DeleteClient(UploadClient):
    def delete(self, items, forced_schemes=None, ignore_availability=False):
        for item in items:
            self._delete_item(item, forced_schemes, ignore_availability)

    def _delete_item(self, item, forced_schemes, ignore_availability):
        logger = self.logger
        dids = [item["did"]]
        files = list(next(self.client.list_replicas(dids))["pfns"].items())
        for file in files:
            pfn = file[0]
            rse = file[1]["rse"]
            force_scheme = None
            for rse_scheme in forced_schemes or []:
                if rse == rse_scheme["rse"]:
                    force_scheme = rse_scheme["scheme"]
            if not self.rses.get(rse):
                rse_settings = self.rses.setdefault(rse, rsemgr.get_rse_info(rse, vo=self.client.vo))
                if not ignore_availability and rse_settings["availability_delete"] != 1:
                    logger(logging.DEBUG, "%s is not available for deletion. No actions have been taken", rse)
                    continue

            # protocol handling and deletion
            rse_settings = self.rses[rse]
            protocols = rsemgr.get_protocols_ordered(rse_settings=rse_settings, operation="delete", scheme=force_scheme)
            protocols.reverse()
            success = False
            while not success and len(protocols):
                protocol = protocols.pop()
                cur_scheme = protocol["scheme"]
                try:
                    protocol_delete = self._create_protocol(rse_settings, "delete", force_scheme=cur_scheme)
                    protocol_delete.delete(pfn)
                    success = True
                except Exception as error:
                    logger(logging.WARNING, "Delete attempt failed")
                    logger(logging.INFO, "Exception: %s", error, exc_info=True)
            logger(logging.DEBUG, "Successfully deleted dataset %s", pfn)


class InPlaceIngestClient(UploadClient):
    def ingest(self, items, summary_file_path=None, traces_copy_out=None, ignore_availability=False, activity=None):
        """
        :param items: List of dictionaries. Each dictionary describing a file to upload. Keys:
            path                  - path of the file that will be uploaded
            path                  - path of the file that will be uploaded
            rse                   - rse expression/name (e.g. 'CERN-PROD_DATADISK') where to upload the file
            did_scope             - Optional: custom did scope (Default: user.<account>)
            did_name              - Optional: custom did name (Default: name of the file)
            dataset_scope         - Optional: custom dataset scope
            dataset_name          - Optional: custom dataset name
            dataset_meta          - Optional: custom metadata for dataset
            impl                  - Optional: name of the protocol implementation to be used to upload this item.
            force_scheme          - Optional: force a specific scheme (if PFN upload this will be overwritten) (Default: None)
            pfn                   - Optional: use a given PFN (this sets no_register to True, and no_register becomes mandatory)
            no_register           - Optional: if True, the file will not be registered in the rucio catalogue
            register_after_upload - Optional: if True, the file will be registered after successful upload
            lifetime              - Optional: the lifetime of the file after it was uploaded
            transfer_timeout      - Optional: time after the upload will be aborted
            guid                  - Optional: guid of the file
            recursive             - Optional: if set, parses the folder structure recursively into collections
        :param summary_file_path: Optional: a path where a summary in form of a json file will be stored
        :param traces_copy_out: reference to an external list, where the traces should be uploaded
        :param ignore_availability: ignore the availability of a RSE
        :param activity: the activity set to the rule if no dataset is specified

        :returns: 0 on success

        :raises InputValidationError: if any input arguments are in a wrong format
        :raises RSEWriteBlocked: if a given RSE is not available for writing
        :raises NoFilesUploaded: if no files were successfully uploaded
        :raises NotAllFilesUploaded: if not all files were successfully uploaded
        """
        # helper to get rse from rse_expression:

        logger = self.logger
        self.trace["uuid"] = generate_uuid()

        # check given sources, resolve dirs into files, and collect meta infos
        files = self._collect_and_validate_file_info(items)
        logger(logging.DEBUG, f"Num. of files that upload client is processing: {len(files)}")

        # check if RSE of every file is available for writing
        # and cache rse settings
        registered_dataset_dids = set()
        registered_file_dids = set()
        for file in files:
            rse = file["rse"]
            if not self.rses.get(rse):
                rse_settings = self.rses.setdefault(rse, rsemgr.get_rse_info(rse, vo=self.client.vo))
                if not ignore_availability and rse_settings["availability_write"] != 1:
                    raise RSEWriteBlocked(f"{rse} is not available for writing. No actions have been taken")

            dataset_scope = file.get("dataset_scope")
            dataset_name = file.get("dataset_name")
            file["rse"] = rse
            if dataset_scope and dataset_name:
                dataset_did_str = f"{dataset_scope}:{dataset_name}"
                file["dataset_did_str"] = dataset_did_str
                registered_dataset_dids.add(dataset_did_str)

            registered_file_dids.add(f"{file['did_scope']}:{file['did_name']}")
        wrong_dids = registered_file_dids.intersection(registered_dataset_dids)
        if len(wrong_dids):
            raise InputValidationError(f"DIDs used to address both files and datasets: {wrong_dids}")
        logger(logging.DEBUG, "Input validation done.")

        # clear this set again to ensure that we only try to register datasets once
        registered_dataset_dids = set()
        num_succeeded = 0
        summary = []
        for file in files:
            basename = file["basename"]
            logger(logging.INFO, "Preparing upload for file %s", basename)

            pfn = file.get("pfn")

            trace = copy.deepcopy(self.trace)
            # appending trace to list reference, if the reference exists
            if traces_copy_out is not None:
                traces_copy_out.append(trace)

            rse = file["rse"]
            trace["scope"] = file["did_scope"]
            trace["datasetScope"] = file.get("dataset_scope", "")
            trace["dataset"] = file.get("dataset_name", "")
            trace["remoteSite"] = rse
            trace["filesize"] = file["bytes"]

            file_did = {"scope": file["did_scope"], "name": file["did_name"]}
            dataset_did_str = file.get("dataset_did_str")
            rse_settings = self.rses[rse]
            is_deterministic = rse_settings.get("deterministic", True)
            if not is_deterministic and not pfn:
                logger(logging.ERROR, "PFN has to be defined for NON-DETERMINISTIC RSE.")
                continue
            if pfn and is_deterministic:
                logger(
                    logging.WARNING,
                    "Upload with given pfn implies that no_register is True, except non-deterministic RSEs",
                )
                no_register = True

            self._register_file(
                file, registered_dataset_dids, ignore_availability=ignore_availability, activity=activity
            )

            file["upload_result"] = {0: True, 1: None, "success": True, "pfn": pfn}  # needs to be removed
            num_succeeded += 1
            trace["transferStart"] = time.time()
            trace["transferEnd"] = time.time()
            trace["clientState"] = "DONE"
            file["state"] = "A"
            logger(logging.INFO, "Successfully uploaded file %s", basename)
            self._send_trace(trace)

            if summary_file_path:
                summary.append(copy.deepcopy(file))

            replica_for_api = self._convert_file_for_api(file)
            try:
                self.client.update_replicas_states(rse, files=[replica_for_api])
            except Exception as error:
                logger(logging.ERROR, f"Failed to update replica state for file {basename}")
                logger(logging.DEBUG, f"Details: {str(error)}")

            # add file to dataset if needed
            if dataset_did_str and not no_register:
                try:
                    self.client.attach_dids(file["dataset_scope"], file["dataset_name"], [file_did])
                except Exception as error:
                    logger(logging.WARNING, "Failed to attach file to the dataset")
                    logger(logging.DEBUG, f"Attaching to dataset {str(error)}")

        if num_succeeded == 0:
            raise NoFilesUploaded()
        elif num_succeeded != len(files):
            raise NotAllFilesUploaded()
        return 0



"""
"""

import logging
import os
import threading
import time
from math import inf
from typing import (
    List,
    Optional,
    Tuple,
)

from typing_extensions import NamedTuple

from galaxy.util import (
    nice_size,
    string_as_bool,
)
from galaxy.util.sleeper import Sleeper

log = logging.getLogger(__name__)


ONE_GIGA_BYTE = 1024 * 1024 * 1024


FileListT = List[Tuple[time.struct_time, str, int]]


class CacheTarget(NamedTuple):
    path: str
    size: int  # cache size in gigabytes
    limit: float  # cache limit as a percent

    def fits_in_cache(self, bytes: int) -> bool:
        # if we don't have a positive cache size - interpret it as an unbounded
        # object store
        if not (self.size > 0):
            return True

        if bytes > (self.size * ONE_GIGA_BYTE * self.limit):
            return False
        return True

    @property
    def log_description(self) -> str:
        return f"{self.limit} percent of {self.size} gigabytes"


def check_caches(targets: List[CacheTarget]):
    for target in targets:
        check_cache(target)


def check_cache(cache_target: CacheTarget):
    """Run a step of the cache monitor."""
    total_size, file_list = _get_cache_size_files(cache_target.path)
    # Sort the file list (based on access time)
    file_list.sort()
    # Initiate cleaning once we reach cache_monitor_cache_limit percentage of the defined cache size?
    # Convert GBs to bytes for comparison
    cache_size_in_gb = cache_target.size * ONE_GIGA_BYTE
    cache_limit = cache_size_in_gb * cache_target.limit
    if total_size > cache_limit:
        log.debug(
            "Initiating cache cleaning: current cache size: %s; clean until smaller than: %s",
            nice_size(total_size),
            nice_size(cache_limit),
        )
        # How much to delete? If simply deleting up to the cache-10% limit,
        # is likely to be deleting frequently and may run the risk of hitting
        # the limit - maybe delete additional #%?
        # For now, delete enough to leave at least 10% of the total cache free
        delete_this_much = total_size - cache_limit
        _clean_cache(file_list, delete_this_much)


def reset_cache(cache_target: CacheTarget):
    _, file_list = _get_cache_size_files(cache_target.path)
    _clean_cache(file_list, inf)


def _clean_cache(file_list: FileListT, delete_this_much: float) -> None:
    """Keep deleting files from the file_list until the size of the deleted
    files is greater than the value in delete_this_much parameter.
    :param file_list: List of candidate files that can be deleted. This method
        will start deleting files from the beginning of the list so the list
        should be sorted accordingly. The list must contains 3-element tuples,
        positioned as follows: position 0 holds file last accessed timestamp
        (as time.struct_time), position 1 holds file path, and position 2 has
        file size (e.g., (<access time>, /mnt/data/dataset_1.dat), 472394)
    :param delete_this_much: Total size of files, in bytes, that should be deleted.
    """
    # Keep deleting datasets from file_list until deleted_amount does not
    # exceed delete_this_much; start deleting from the front of the file list,
    # which assumes the oldest files come first on the list.
    deleted_amount = 0
    for entry in file_list:
        if deleted_amount < delete_this_much:
            deleted_amount += entry[2]
            os.remove(entry[1])
        else:
            log.debug("Cache cleaning done. Total space freed: %s", nice_size(deleted_amount))
            return


def _get_cache_size_files(cache_path) -> Tuple[int, FileListT]:
    """Returns cache size and cache files.

    For each file, we get last access time, file path, and file size.
    """
    cache_size = 0
    file_list = []

    for dirpath, _, filenames in os.walk(cache_path):
        for filename in filenames:
            file_path = os.path.join(dirpath, filename)
            file_size = os.path.getsize(file_path)
            cache_size += file_size
            # Get the time given file was last accessed
            last_access_time = time.localtime(os.stat(file_path)[7])
            # Compose a tuple of the access time and the file path
            file_tuple = last_access_time, file_path, file_size
            file_list.append(file_tuple)
    return cache_size, file_list


def parse_caching_config_dict_from_xml(config_xml):
    cache_els = config_xml.findall("cache")
    if len(cache_els) > 0:
        c_xml = config_xml.findall("cache")[0]
        cache_size = float(c_xml.get("size", -1))
        staging_path = c_xml.get("path", None)
        monitor = c_xml.get("monitor", "auto")
        cache_updated_data = string_as_bool(c_xml.get("cache_updated_data", "True"))

        cache_dict = {
            "size": cache_size,
            "path": staging_path,
            "monitor": monitor,
            "cache_updated_data": cache_updated_data,
        }
    else:
        cache_dict = {}
    return cache_dict


def configured_cache_size(config, config_dict) -> int:
    cache_config_dict = config_dict.get("cache") or {}
    cache_size = cache_config_dict.get("size") or config.object_store_cache_size
    if cache_size != -1:
        # Convert admin-set GBs to bytes internally for quick comparison
        cache_size = cache_size * ONE_GIGA_BYTE
    return cache_size


def enable_cache_monitor(config, config_dict) -> Tuple[bool, int]:
    cache_config_dict = config_dict.get("cache") or {}
    default_interval = getattr(config, "object_store_cache_monitor_interval", 600)
    interval = cache_config_dict.get("monitor_interval") or default_interval

    disable_process_management = getattr(config, "disable_process_management", None)
    if disable_process_management is True:
        return False, interval

    if config_dict.get("enable_cache_monitor", False) is False:
        return False, interval

    default_cache_driver = getattr(config, "object_store_cache_monitor_driver", "auto")
    monitor = cache_config_dict.get("monitor", default_cache_driver)
    if monitor == "auto":
        monitor = "celery" if getattr(config, "enable_celery_tasks", False) else "inprocess"

    return monitor == "inprocess", interval


class InProcessCacheMonitor:
    def __init__(self, cache_target: CacheTarget, interval: int = 30, initial_sleep: Optional[int] = 2):
        # This Event object is initialized to False
        # It is set to True in shutdown(), causing
        # the cache monitor thread to return/terminate
        self.stop_cache_monitor_event = threading.Event()
        # Helper for interruptable sleep
        self.sleeper = Sleeper()

        self.cache_target = cache_target
        self.interval = interval
        self.initial_sleep = initial_sleep

        self.cache_monitor_thread = threading.Thread(
            target=self._cache_monitor,
            name="CacheMonitorThread",
        )
        self.cache_monitor_thread.start()

    def _cache_monitor(self):
        if self.initial_sleep is not None:
            time.sleep(
                self.initial_sleep
            )  # startup sleep hack - probably originally implemented to prevent contention at app startup
        while not self.stop_cache_monitor_event.is_set():
            check_cache(self.cache_target)
            self.sleeper.sleep(self.interval)

    def shutdown(self):
        # Set the event object so the cache monitor thread terminates
        self.stop_cache_monitor_event.set()

        # wake up from sleeping
        self.sleeper.wake()

        # Wait for the cache monitor thread to join before ending
        self.cache_monitor_thread.join(5)



import multiprocessing
import os
import subprocess

from galaxy.util import (
    umask_fix_perms,
    which,
)


def fix_permissions(config, rel_path: str):
    """Set permissions on rel_path"""
    for basedir, _, files in os.walk(rel_path):
        umask_fix_perms(basedir, config.umask, 0o777, config.gid)
        for filename in files:
            path = os.path.join(basedir, filename)
            # Ignore symlinks
            if os.path.islink(path):
                continue
            umask_fix_perms(path, config.umask, 0o666, config.gid)


class UsesAxel:
    use_axel: bool

    def _init_axel(self) -> None:
        if which("axel"):
            self.use_axel = True
        else:
            self.use_axel = False

    def _axel_download(self, url: str, path: str):
        ncores = multiprocessing.cpu_count()
        ret_code = subprocess.call(["axel", "-a", "-o", path, "-n", str(ncores), url])
        return ret_code == 0



from . import BaseObjectStore

try:
    from pulsar.client.manager import ObjectStoreClientManager
except ImportError:
    ObjectStoreClientManager = None


class PulsarObjectStore(BaseObjectStore):
    """
    Object store implementation that delegates to a remote Pulsar server.

    This may be more aspirational than practical for now, it would be good to
    Galaxy to a point that a handler thread could be setup that doesn't attempt
    to access the disk files returned by a (this) object store - just passing
    them along to the Pulsar unmodified. That modification - along with this
    implementation and Pulsar job destinations would then allow Galaxy to fully
    manage jobs on remote servers with completely different mount points.

    This implementation should be considered beta and may be dropped from
    Galaxy at some future point or significantly modified.
    """

    def __init__(self, config, config_xml):
        self.pulsar_client = self.__build_pulsar_client(config_xml)

    def _exists(self, obj, **kwds) -> bool:
        return self.pulsar_client.exists(**self.__build_kwds(obj, **kwds))

    def file_ready(self, obj, **kwds):
        return self.pulsar_client.file_ready(**self.__build_kwds(obj, **kwds))

    def _create(self, obj, **kwds):
        return self.pulsar_client.create(**self.__build_kwds(obj, **kwds))

    def _empty(self, obj, **kwds) -> bool:
        return self.pulsar_client.empty(**self.__build_kwds(obj, **kwds))

    def _size(self, obj, **kwds) -> int:
        return self.pulsar_client.size(**self.__build_kwds(obj, **kwds))

    def _delete(self, obj, **kwds) -> bool:
        return self.pulsar_client.delete(**self.__build_kwds(obj, **kwds))

    # TODO: Optimize get_data.
    def _get_data(self, obj, **kwds):
        return self.pulsar_client.get_data(**self.__build_kwds(obj, **kwds))

    def _get_filename(self, obj, **kwds) -> str:
        return self.pulsar_client.get_filename(**self.__build_kwds(obj, **kwds))

    def _update_from_file(self, obj, **kwds) -> None:
        return self.pulsar_client.update_from_file(**self.__build_kwds(obj, **kwds))

    def _get_store_usage_percent(self):
        return self.pulsar_client.get_store_usage_percent()

    def _get_object_url(self, obj, extra_dir=None, extra_dir_at_root=False, alt_name=None):
        return None

    def __build_kwds(self, obj, **kwds):
        kwds["object_id"] = obj.id
        return kwds

    def __build_pulsar_client(self, config_xml):
        if ObjectStoreClientManager is None:
            raise Exception("Pulsar client code not available, cannot use this module.")
        url = config_xml.get("url")
        private_token = config_xml.get("private_token", None)
        transport = config_xml.get("transport", None)
        manager_options = dict(transport=transport)
        client_options = dict(url=url, private_token=private_token)
        pulsar_client = ObjectStoreClientManager(**manager_options).get_client(client_options)
        return pulsar_client

    def shutdown(self):
        pass



from .manager import ConfiguredObjectStoreTemplates
from .models import (
    ObjectStoreConfiguration,
    ObjectStoreTemplate,
    ObjectStoreTemplateSummaries,
    ObjectStoreTemplateType,
    template_to_configuration,
)

__all__ = (
    "ConfiguredObjectStoreTemplates",
    "ObjectStoreConfiguration",
    "ObjectStoreTemplate",
    "ObjectStoreTemplateSummaries",
    "ObjectStoreTemplateType",
    "template_to_configuration",
)



from typing import (
    Any,
    Dict,
    List,
    Optional,
    Set,
    Union,
)

from typing_extensions import (
    Literal,
    NotRequired,
    TypedDict,
)

BadgeSourceT = Literal["admin", "galaxy"]
# Badges that can be explicitly set by admins.
AdminBadgeT = Literal[
    "faster",
    "slower",
    "short_term",
    "backed_up",
    "not_backed_up",
    "more_secure",
    "less_secure",
    "more_stable",
    "less_stable",
]

# All badges - so AdminBadgeT plus Galaxy specifiable badges.
BadgeT = Union[
    AdminBadgeT,
    Literal[
        "cloud",
        "quota",
        "no_quota",
        "restricted",
        "user_defined",
    ],
]


class BadgeSpecDict(TypedDict):
    """Describe badges that can be set by Galaxy admins."""

    type: AdminBadgeT
    conflicts: List[AdminBadgeT]


BADGE_SPECIFICATION: List[BadgeSpecDict] = [
    {"type": "faster", "conflicts": ["slower"]},
    {"type": "slower", "conflicts": ["faster"]},
    {"type": "short_term", "conflicts": []},
    {"type": "backed_up", "conflicts": ["not_backed_up"]},
    {"type": "not_backed_up", "conflicts": ["backed_up"]},
    {"type": "more_secure", "conflicts": ["less_secure"]},
    {"type": "less_secure", "conflicts": ["more_secure"]},
    {"type": "more_stable", "conflicts": ["less_stable"]},
    {"type": "less_stable", "conflicts": ["more_stable"]},
]

KNOWN_BADGE_TYPES: List[AdminBadgeT] = [s["type"] for s in BADGE_SPECIFICATION]
BADGE_SPECIFICATION_BY_TYPE: Dict[AdminBadgeT, BadgeSpecDict] = {s["type"]: s for s in BADGE_SPECIFICATION}


class BadgeDict(TypedDict):
    type: BadgeT
    message: Optional[str]
    source: BadgeSourceT


class StoredBadgeDict(TypedDict):
    type: AdminBadgeT
    message: NotRequired[Optional[str]]


def read_badges(config_dict: Dict[str, Any]) -> List[StoredBadgeDict]:
    raw_badges = config_dict.get("badges") or []
    badges: List[StoredBadgeDict] = []
    badge_types: Set[str] = set()
    badge_conflicts: Dict[str, str] = {}
    for badge in raw_badges:
        # when recovering serialized badges, skip ones that are set by Galaxy
        badge_source = badge.get("source")
        if badge_source and badge_source != "admin":
            continue

        assert "type" in badge
        badge_type = badge["type"]
        if badge_type not in KNOWN_BADGE_TYPES:
            raise Exception(f"badge_type {badge_type} unimplemented/unknown {badge}")
        message = badge.get("message", None)
        badges.append({"type": badge_type, "message": message})
        badge_types.add(badge_type)
        if badge_type in badge_conflicts:
            conflicting_badge_type = badge_conflicts[badge_type]
            raise Exception(
                f"Conflicting badge to [{badge_type}] defined on the object store [{conflicting_badge_type}]."
            )
        conflicts = BADGE_SPECIFICATION_BY_TYPE[badge_type]["conflicts"]
        for conflict in conflicts:
            badge_conflicts[conflict] = badge_type
    return badges


def serialize_badges(
    stored_badges: List[StoredBadgeDict], quota_enabled: bool, private: bool, user_defined: bool, cloud: bool
) -> List[BadgeDict]:
    """Produce blended, unified list of badges for target object store entity.

    Combine more free form admin information stored about badges with Galaxy tracked
    information (quota and access restriction information) to produce a unified list
    of badges to be consumed via the API.
    """
    badge_dicts: List[BadgeDict] = []
    for badge in stored_badges:
        badge_dict: BadgeDict = {
            "source": "admin",
            "type": badge["type"],
            "message": badge.get("message"),
        }
        badge_dicts.append(badge_dict)

    quota_badge_dict: BadgeDict
    if quota_enabled:
        quota_badge_dict = {
            "type": "quota",
            "message": None,
            "source": "galaxy",
        }
    else:
        quota_badge_dict = {
            "type": "no_quota",
            "message": None,
            "source": "galaxy",
        }
    badge_dicts.append(quota_badge_dict)
    if private:
        restricted_badge_dict: BadgeDict = {
            "type": "restricted",
            "message": None,
            "source": "galaxy",
        }
        badge_dicts.append(restricted_badge_dict)
    if user_defined:
        user_defined_badge_dict: BadgeDict = {
            "type": "user_defined",
            "message": None,
            "source": "galaxy",
        }
        badge_dicts.append(user_defined_badge_dict)
    if cloud:
        cloud_badge_dict: BadgeDict = {
            "type": "cloud",
            "message": None,
            "source": "galaxy",
        }
        badge_dicts.append(cloud_badge_dict)
    return badge_dicts



#!/usr/bin/env python
"""
Split large file into multiple pieces for upload to S3.
This parallelizes the task over available cores using multiprocessing.
Code mostly taken form CloudBioLinux.
"""

import glob
import os
import subprocess
import threading

try:
    import boto
    from boto.s3.connection import S3Connection
except ImportError:
    boto = None  # type: ignore[assignment]


def mp_from_ids(s3server, mp_id, mp_keyname, mp_bucketname):
    """Get the multipart upload from the bucket and multipart IDs.

    This allows us to reconstitute a connection to the upload
    from within multiprocessing functions.
    """
    if s3server["host"]:
        conn = boto.connect_s3(
            aws_access_key_id=s3server["access_key"],
            aws_secret_access_key=s3server["secret_key"],
            is_secure=s3server["is_secure"],
            host=s3server["host"],
            port=s3server["port"],
            calling_format=boto.s3.connection.OrdinaryCallingFormat(),
            path=s3server["conn_path"],
        )
    else:
        if s3server["access_key"]:
            conn = S3Connection(s3server["access_key"], s3server["secret_key"])
        else:
            conn = S3Connection()

    bucket = conn.lookup(mp_bucketname)
    mp = boto.s3.multipart.MultiPartUpload(bucket)
    mp.key_name = mp_keyname
    mp.id = mp_id
    return mp


def transfer_part(s3server, mp_id, mp_keyname, mp_bucketname, i, part):
    """Transfer a part of a multipart upload. Designed to be run in parallel."""
    mp = mp_from_ids(s3server, mp_id, mp_keyname, mp_bucketname)
    with open(part, "rb") as t_handle:
        mp.upload_part_from_file(t_handle, i + 1)
    os.remove(part)


def multipart_upload(s3server, bucket, s3_key_name, tarball, mb_size):
    """Upload large files using Amazon's multipart upload functionality."""

    def split_file(in_file, mb_size, split_num=5):
        prefix = os.path.join(os.path.dirname(in_file), f"{os.path.basename(s3_key_name)}S3PART")
        max_chunk = s3server["max_chunk_size"]
        # Split chunks so they are 5MB < chunk < 250MB(max_chunk_size)
        split_size = int(max(min(mb_size / (split_num * 2.0), max_chunk), 5))
        if not os.path.exists(f"{prefix}aa"):
            cl = ["split", f"-b{split_size}m", in_file, prefix]
            subprocess.check_call(cl)
        return sorted(glob.glob(f"{prefix}*"))

    mp = bucket.initiate_multipart_upload(s3_key_name, reduced_redundancy=s3server["use_rr"])

    for i, part in enumerate(split_file(tarball, mb_size)):
        t = threading.Thread(target=transfer_part, args=(s3server, mp.id, mp.key_name, mp.bucket_name, i, part))
        t.start()
        t.join()

    mp.complete_upload()



import logging
import os
import shutil
from datetime import datetime
from typing import (
    Any,
    Dict,
    Optional,
)

from galaxy.exceptions import (
    ObjectInvalid,
    ObjectNotFound,
)
from galaxy.objectstore import ConcreteObjectStore
from galaxy.util import (
    directory_hash_id,
    unlink,
)
from galaxy.util.path import safe_relpath
from ._util import fix_permissions
from .caching import (
    CacheTarget,
    InProcessCacheMonitor,
)

log = logging.getLogger(__name__)


class CachingConcreteObjectStore(ConcreteObjectStore):
    staging_path: str
    extra_dirs: Dict[str, str]
    config: Any
    cache_updated_data: bool
    enable_cache_monitor: bool
    cache_size: int
    cache_monitor: Optional[InProcessCacheMonitor] = None
    cache_monitor_interval: int

    def _ensure_staging_path_writable(self):
        staging_path = self.staging_path
        if not os.path.exists(staging_path):
            os.makedirs(staging_path, exist_ok=True)
            if not os.path.exists(staging_path):
                raise Exception(f"Caching object store created with path '{staging_path}' that does not exist")

        if not os.access(staging_path, os.R_OK):
            raise Exception(f"Caching object store created with path '{staging_path}' that does not readable")
        if not os.access(staging_path, os.W_OK):
            raise Exception(f"Caching object store created with path '{staging_path}' that does not writable")

    def _construct_path(
        self,
        obj,
        base_dir=None,
        dir_only=None,
        extra_dir=None,
        extra_dir_at_root=False,
        alt_name=None,
        obj_dir: bool = False,
        in_cache: bool = False,
    ) -> str:
        # extra_dir should never be constructed from provided data but just
        # make sure there are no shenannigans afoot
        if extra_dir and extra_dir != os.path.normpath(extra_dir):
            log.warning("extra_dir is not normalized: %s", extra_dir)
            raise ObjectInvalid("The requested object is invalid")
        # ensure that any parent directory references in alt_name would not
        # result in a path not contained in the directory path constructed here
        if alt_name:
            if not safe_relpath(alt_name):
                log.warning("alt_name would locate path outside dir: %s", alt_name)
                raise ObjectInvalid("The requested object is invalid")
            # alt_name can contain parent directory references, but S3 will not
            # follow them, so if they are valid we normalize them out
            alt_name = os.path.normpath(alt_name)

        object_id = self._get_object_id(obj)
        rel_path = os.path.join(*directory_hash_id(object_id))

        if extra_dir is not None:
            if extra_dir_at_root:
                rel_path = os.path.join(extra_dir, rel_path)
            else:
                rel_path = os.path.join(rel_path, extra_dir)

        # for JOB_WORK directory
        if obj_dir:
            rel_path = os.path.join(rel_path, str(object_id))
        if base_dir:
            base = self.extra_dirs.get(base_dir)
            assert base
            return os.path.join(base, rel_path)

        # This is how the remote file stores represent folders
        rel_path = f"{rel_path}/"

        if not dir_only:
            rel_path = os.path.join(rel_path, alt_name if alt_name else f"dataset_{object_id}.dat")

        if in_cache:
            return self._get_cache_path(rel_path)

        return rel_path

    def _get_cache_path(self, rel_path: str) -> str:
        return os.path.abspath(os.path.join(self.staging_path, rel_path))

    def _in_cache(self, rel_path: str) -> bool:
        """Check if the given dataset is in the local cache and return True if so."""
        cache_path = self._get_cache_path(rel_path)
        return os.path.exists(cache_path)

    def _pull_into_cache(self, rel_path, **kwargs) -> bool:
        # Ensure the cache directory structure exists (e.g., dataset_#_files/)
        rel_path_dir = os.path.dirname(rel_path)
        if not os.path.exists(self._get_cache_path(rel_path_dir)):
            os.makedirs(self._get_cache_path(rel_path_dir), exist_ok=True)
        # Now pull in the file
        file_ok = self._download(rel_path)
        if file_ok:
            fix_permissions(self.config, self._get_cache_path(rel_path_dir))
        else:
            unlink(self._get_cache_path(rel_path), ignore_errors=True)
        return file_ok

    def _get_data(self, obj, start=0, count=-1, **kwargs):
        rel_path = self._construct_path(obj, **kwargs)
        # Check cache first and get file if not there
        if not self._in_cache(rel_path):
            self._pull_into_cache(rel_path, **kwargs)
        # Read the file content from cache
        data_file = open(self._get_cache_path(rel_path))
        data_file.seek(start)
        content = data_file.read(count)
        data_file.close()
        return content

    def _exists(self, obj, **kwargs) -> bool:
        in_cache = exists_remotely = False
        rel_path = self._construct_path(obj, **kwargs)
        dir_only = kwargs.get("dir_only", False)
        base_dir = kwargs.get("base_dir", None)

        # check job work directory stuff early to skip API hits.
        if dir_only and base_dir:
            if not os.path.exists(rel_path):
                os.makedirs(rel_path, exist_ok=True)
            return True

        in_cache = self._in_cache(rel_path)
        exists_remotely = self._exists_remotely(rel_path)
        dir_only = kwargs.get("dir_only", False)
        base_dir = kwargs.get("base_dir", None)
        if dir_only:
            if in_cache or exists_remotely:
                return True
            else:
                return False

        # TODO: Sync should probably not be done here. Add this to an async upload stack?
        if in_cache and not exists_remotely:
            self._push_to_storage(rel_path, source_file=self._get_cache_path(rel_path))
            return True
        elif exists_remotely:
            return True
        else:
            return False

    def _create(self, obj, **kwargs):
        if not self._exists(obj, **kwargs):
            # Pull out locally used fields
            extra_dir = kwargs.get("extra_dir", None)
            extra_dir_at_root = kwargs.get("extra_dir_at_root", False)
            dir_only = kwargs.get("dir_only", False)
            alt_name = kwargs.get("alt_name", None)

            # Construct hashed path
            rel_path = os.path.join(*directory_hash_id(self._get_object_id(obj)))

            # Optionally append extra_dir
            if extra_dir is not None:
                if extra_dir_at_root:
                    rel_path = os.path.join(extra_dir, rel_path)
                else:
                    rel_path = os.path.join(rel_path, extra_dir)

            # Create given directory in cache
            cache_dir = os.path.join(self.staging_path, rel_path)
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir, exist_ok=True)

            # If instructed, create the dataset in cache & in S3
            if not dir_only:
                rel_path = os.path.join(rel_path, alt_name if alt_name else f"dataset_{self._get_object_id(obj)}.dat")
                open(os.path.join(self.staging_path, rel_path), "w").close()
                self._push_to_storage(rel_path, from_string="")
        return self

    def _caching_allowed(self, rel_path: str, remote_size: Optional[int] = None) -> bool:
        if remote_size is None:
            remote_size = self._get_remote_size(rel_path)
        if not self.cache_target.fits_in_cache(remote_size):
            log.critical(
                "File %s is larger (%s bytes) than the configured cache allows (%s). Cannot download.",
                rel_path,
                remote_size,
                self.cache_target.log_description,
            )
            return False
        return True

    def _push_to_storage(self, rel_path, source_file=None, from_string=None):
        source_file = source_file or self._get_cache_path(rel_path)
        if from_string is None and not os.path.exists(source_file):
            log.error(
                "Tried updating remote path '%s' from source file '%s', but source file does not exist.",
                rel_path,
                source_file,
            )
            return False

        if from_string is None and os.path.getsize(source_file) == 0:
            log.debug(
                "Wanted to push file '%s' to remote path '%s' but its size is 0; skipping.", source_file, rel_path
            )
            return True

        if from_string is not None:
            return self._push_string_to_path(rel_path, from_string)
        else:
            start_time = datetime.now()
            log.debug(
                "Pushing cache file '%s' of size %s bytes to '%s'",
                source_file,
                os.path.getsize(source_file),
                rel_path,
            )
            success = self._push_file_to_path(rel_path, source_file)
            end_time = datetime.now()
            log.debug(
                "Pushed cache file '%s' to blob '%s' (%s bytes transferred in %s sec)",
                source_file,
                rel_path,
                os.path.getsize(source_file),
                end_time - start_time,
            )
            return success

    def _empty(self, obj, **kwargs) -> bool:
        if self._exists(obj, **kwargs):
            return self._size(obj, **kwargs) == 0
        else:
            raise ObjectNotFound(f"objectstore.empty, object does not exist: {obj}, kwargs: {kwargs}")

    def _get_size_in_cache(self, rel_path):
        return os.path.getsize(self._get_cache_path(rel_path))

    def _size(self, obj, **kwargs) -> int:
        rel_path = self._construct_path(obj, **kwargs)
        if self._in_cache(rel_path):
            try:
                return self._get_size_in_cache(rel_path)
            except OSError as ex:
                log.info("Could not get size of file '%s' in local cache, will try Azure. Error: %s", rel_path, ex)
        elif self._exists_remotely(rel_path):
            return self._get_remote_size(rel_path)
        log.warning("Did not find dataset '%s', returning 0 for size", rel_path)
        return 0

    def _get_filename(self, obj, sync_cache: bool = True, **kwargs) -> str:
        base_dir = kwargs.get("base_dir", None)
        dir_only = kwargs.get("dir_only", False)
        obj_dir = kwargs.get("obj_dir", False)

        rel_path = self._construct_path(obj, **kwargs)

        # for JOB_WORK directory
        if base_dir and dir_only and obj_dir:
            return os.path.abspath(rel_path)

        cache_path = self._get_cache_path(rel_path)
        if not sync_cache:
            return cache_path

        # Check if the file exists in the cache first, always pull if file size in cache is zero
        # For dir_only - the cache cleaning may have left empty directories so I think we need to
        # always resync the cache. Gotta make sure we're being judicious in out data.extra_files_path
        # calls I think.
        if not dir_only and self._in_cache(rel_path) and os.path.getsize(self._get_cache_path(rel_path)) > 0:
            return cache_path

        # Check if the file exists in persistent storage and, if it does, pull it into cache
        elif self._exists(obj, **kwargs):
            if dir_only:
                self._download_directory_into_cache(rel_path, cache_path)
                return cache_path
            else:
                if self._pull_into_cache(rel_path, **kwargs):
                    return cache_path
        raise ObjectNotFound(f"objectstore.get_filename, no cache_path: {obj}, kwargs: {kwargs}")

    def _download_directory_into_cache(self, rel_path, cache_path):
        # pithos & irods never did this prior to refactoring so I am assuming
        # there is just operations that fail with these object stores.
        # As part of the refactoring that resulted in this method
        # https://github.com/galaxyproject/galaxy/pull/18117 I wrote test
        # cases and I verified the other object stores that didn't implement
        # this had issues - I implemented this new functionality in the
        # Azure and Cloud object stores to fix those object stores. New
        # object stores should definitely override this.
        pass

    def _delete(self, obj, entire_dir: bool = False, **kwargs) -> bool:
        rel_path = self._construct_path(obj, **kwargs)
        extra_dir = kwargs.get("extra_dir", None)
        base_dir = kwargs.get("base_dir", None)
        dir_only = kwargs.get("dir_only", False)
        obj_dir = kwargs.get("obj_dir", False)
        try:
            # Remove temporary data in JOB_WORK directory
            if base_dir and dir_only and obj_dir:
                shutil.rmtree(os.path.abspath(rel_path))
                return True

            # For the case of extra_files, because we don't have a reference to
            # individual files/keys we need to remove the entire directory structure
            # with all the files in it. This is easy for the local file system,
            # but requires iterating through each individual key in S3 and deleing it.
            if entire_dir and extra_dir:
                shutil.rmtree(self._get_cache_path(rel_path), ignore_errors=True)
                return self._delete_remote_all(rel_path)
            else:
                # Delete from cache first
                unlink(self._get_cache_path(rel_path), ignore_errors=True)
                # Delete from S3 as well
                if self._exists_remotely(rel_path):
                    return self._delete_existing_remote(rel_path)
        except OSError:
            log.exception("%s delete error", self._get_filename(obj, **kwargs))
        return False

    def _update_from_file(
        self, obj, file_name=None, create: bool = False, preserve_symlinks: bool = False, **kwargs
    ) -> None:
        if create:
            self._create(obj, **kwargs)

        if self._exists(obj, **kwargs):
            rel_path = self._construct_path(obj, **kwargs)
            # Chose whether to use the dataset file itself or an alternate file
            if file_name:
                source_file = os.path.abspath(file_name)
                # Copy into cache
                cache_file = self._get_cache_path(rel_path)
                try:
                    if source_file != cache_file and self.cache_updated_data:
                        # FIXME? Should this be a `move`?
                        shutil.copy2(source_file, cache_file)
                    fix_permissions(self.config, cache_file)
                except OSError:
                    log.exception("Trouble copying source file '%s' to cache '%s'", source_file, cache_file)
            else:
                source_file = self._get_cache_path(rel_path)

            self._push_to_storage(rel_path, source_file)

        else:
            raise ObjectNotFound(
                f"objectstore.update_from_file, object does not exist: {str(obj)}, kwargs: {str(kwargs)}"
            )

    @property
    def cache_target(self) -> CacheTarget:
        return CacheTarget(
            self.staging_path,
            self.cache_size,
            0.9,
        )

    def _shutdown_cache_monitor(self) -> None:
        self.cache_monitor and self.cache_monitor.shutdown()

    def _start_cache_monitor_if_needed(self):
        if self.enable_cache_monitor:
            self.cache_monitor = InProcessCacheMonitor(self.cache_target, self.cache_monitor_interval)

    def _get_remote_size(self, rel_path: str) -> int:
        raise NotImplementedError()

    def _exists_remotely(self, rel_path: str) -> bool:
        raise NotImplementedError()

    def _download(self, rel_path: str) -> bool:
        raise NotImplementedError()

    # Do not need to override these if instead replacing _delete
    def _delete_existing_remote(self, rel_path) -> bool:
        raise NotImplementedError()

    def _delete_remote_all(self, rel_path) -> bool:
        raise NotImplementedError()

    # Do not need to override these if instead replacing _push_to_storage
    def _push_string_to_path(self, rel_path: str, from_string: str) -> bool:
        raise NotImplementedError()

    def _push_file_to_path(self, rel_path: str, target_file: str) -> bool:
        raise NotImplementedError()



# This work originally developed by Stavros Sachtouris <saxtouri@grnet.gr>
# as part of the effort committed GRNET S.A. (Greek Research and Technology
# Network) in the context of the OpenMinTeD project (openminted.eu)

import logging
import os

try:
    from kamaki.clients import (
        astakos,
        Client as KamakiClient,
        ClientError,
        pithos,
        utils,
    )
except ImportError:
    KamakiClient = None

from galaxy.util import directory_hash_id
from ._caching_base import CachingConcreteObjectStore

NO_KAMAKI_ERROR_MESSAGE = (
    "ObjectStore configured, but no kamaki.clients dependency available."
    "Please install and properly configure kamaki.clients or modify Object "
    "Store configuration."
)

log = logging.getLogger(__name__)


def parse_config_xml(config_xml):
    """Parse and validate config_xml, return dict for convenience
    :param config_xml: (lxml.etree.Element) root of XML subtree
    :returns: (dict) according to syntax
    :raises: various XML parse errors
    """
    r = {}
    try:
        for tag, required_attrs, optional_attrs in (
            (
                "auth",
                (
                    "url",
                    "token",
                ),
                (
                    "ca_certs",
                    "ignore_ssl",
                ),
            ),
            ("container", ("name",), ("project",)),
        ):
            element = config_xml.findall(tag)[0]
            required = tuple((k, element.get(k)) for k in required_attrs)
            for k, v in required:
                if not v:
                    msg = f"No value for {tag}:{k} in XML tree"
                    log.error(msg)
                    raise Exception(msg)
            optional = tuple((k, element.get(k)) for k in optional_attrs)
            r[tag] = dict(required + optional)

        # Extract extra_dir
        tag, attrs = "extra_dir", ("type", "path")
        extra_dirs = config_xml.findall(tag)
        if not extra_dirs:
            msg = f"No {tag} element in XML tree"
            log.error(msg)
            raise Exception(msg)
        r["extra_dirs"] = [{k: e.get(k) for k in attrs} for e in extra_dirs]
        r["private"] = CachingConcreteObjectStore.parse_private_from_config_xml(config_xml)
        if "job_work" not in (d["type"] for d in r["extra_dirs"]):
            msg = f'No value for {tag}:type="job_work" in XML tree'
            log.error(msg)
            raise Exception(msg)
    except Exception:
        log.exception("Malformed PithosObjectStore Configuration XML, unable to continue")
        raise
    return r


class PithosObjectStore(CachingConcreteObjectStore):
    """
    Object store that stores objects as items in a Pithos+ container.
    Cache is ignored for the time being.
    """

    store_type = "pithos"

    def __init__(self, config, config_dict):
        super().__init__(config, config_dict)
        self.staging_path = self.config.file_path
        log.info("Parse config_xml for pithos object store")
        self.config_dict = config_dict

        self._initialize()

    def _initialize(self):
        if KamakiClient is None:
            raise Exception(NO_KAMAKI_ERROR_MESSAGE)

        self._ensure_staging_path_writable()
        log.info("Authenticate Synnefo account")
        self._authenticate()
        log.info("Initialize Pithos+ client")
        self._init_pithos()

    @classmethod
    def parse_xml(clazz, config_xml):
        return parse_config_xml(config_xml)

    def to_dict(self):
        as_dict = super().to_dict()
        as_dict.update(self.config_dict)
        return as_dict

    def _authenticate(self):
        auth = self.config_dict["auth"]
        url, token = auth["url"], auth["token"]
        ca_certs = auth.get("ca_certs")
        if ca_certs:
            utils.https.patch_with_certs(ca_certs)
        elif auth.get("ignore_ssl").lower() in ("true", "yes", "on"):
            utils.https.patch_ignore_ssl()
        self.astakos = astakos.AstakosClient(url, token)

    def _init_pithos(self):
        uuid, token = self.astakos.user_term("id"), self.astakos.token
        service_type = pithos.PithosClient.service_type
        pithos_url = self.astakos.get_endpoint_url(service_type)
        container = self.config_dict["container"]["name"]
        self.pithos = pithos.PithosClient(pithos_url, token, uuid, container)

        # Create container if not exist, or reassign to named project
        project = self.config_dict["container"].get("project", None)
        try:
            c = self.pithos.get_container_info()
        except ClientError as ce:
            if ce.status not in (404,):
                raise
            c = self.pithos.create_container(project_id=project)
            return
        if project and c.get("x-container-policy-project") != project:
            self.pithos.reassign_container(project)

    def _download(self, rel_path):
        local_destination = self._get_cache_path(rel_path)
        self.pithos.download_object(rel_path, local_destination)

    # No need to overwrite "shutdown"

    def _exists(self, obj, **kwargs) -> bool:
        """Check if file exists, fix if file in cache and not on Pithos+
        :returns: weather the file exists remotely or in cache
        """
        path = self._construct_path(obj, **kwargs)
        try:
            self.pithos.get_object_info(path)
            return True
        except ClientError as ce:
            if ce.status not in (404,):
                raise

        in_cache = self._in_cache(path)
        dir_only = kwargs.get("dir_only", False)
        if dir_only:
            base_dir = kwargs.get("base_dir", None)
            if in_cache:
                return True
            elif base_dir:  # for JOB_WORK directory
                if not os.path.exists(path):
                    os.makedirs(path, exist_ok=True)
                return True
            return False

        if in_cache:
            cache_path = self._get_cache_path(path)
            # Maybe the upload should have happened in some thread elsewhere?
            with open(cache_path) as f:
                self.pithos.upload_object(path, f)
            return True
        return False

    def _create(self, obj, **kwargs):
        """Touch a file (aka create empty), if it doesn't exist"""
        if not self._exists(obj, **kwargs):
            # Pull out locally used fields
            extra_dir = kwargs.get("extra_dir", None)
            extra_dir_at_root = kwargs.get("extra_dir_at_root", False)
            dir_only = kwargs.get("dir_only", False)
            alt_name = kwargs.get("alt_name", None)

            # Construct hashed path
            rel_path = os.path.join(*directory_hash_id(self._get_object_id(obj)))

            # Optionally append extra_dir
            if extra_dir is not None:
                if extra_dir_at_root:
                    rel_path = os.path.join(extra_dir, rel_path)
                else:
                    rel_path = os.path.join(rel_path, extra_dir)

            # Create given directory in cache
            cache_dir = os.path.join(self.staging_path, rel_path)
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir, exist_ok=True)

            if dir_only:
                self.pithos.upload_from_string(rel_path, "", content_type="application/directory")
            else:
                rel_path = os.path.join(rel_path, alt_name if alt_name else f"dataset_{self._get_object_id(obj)}.dat")
                new_file = os.path.join(self.staging_path, rel_path)
                open(new_file, "w").close()
                self.pithos.upload_from_string(rel_path, "")
        return self

    def _get_remote_size(self, path):
        try:
            file = self.pithos.get_object_info(path)
        except ClientError as ce:
            if ce.status not in (404,):
                raise
            return 0
        return int(file["content-length"])

    def _delete_remote_all(self, path: str) -> bool:
        try:
            log.debug(f"On Pithos: delete -r {path}/")
            self.pithos.del_object(path, delimiter="/")
            return True
        except ClientError:
            log.exception(f"Could not delete path '{path}' from Pithos")
            return False

    def _delete_existing_remote(self, path: str) -> bool:
        try:
            self.pithos.del_object(path)
            return True
        except ClientError:
            log.exception(f"Could not delete path '{path}' from Pithos")
            return False

    def _get_object_url(self, obj, **kwargs):
        """
        :returns: URL for direct access, None if no object
        """
        if self._exists(obj, **kwargs):
            path = self._construct_path(obj, **kwargs)
            try:
                return self.pithos.publish_object(path)
            except ClientError as ce:
                log.exception(f'Trouble generating URL for dataset "{path}"')
                log.exception(f"Kamaki: {ce}")
        return None

    def _get_store_usage_percent(self):
        """
        :returns: percentage indicating how full the store is
        """
        quotas = self.astakos.get_quotas()
        project = self.config_dict["container"]["project"]
        pithos_quotas = quotas[project]["pithos.diskspace"]
        usage = pithos_quotas["usage"]
        limit = min(pithos_quotas["limit"], pithos_quotas["project_limit"])
        return (100.0 * usage) / limit



"""
Object Store plugin for the Onedata Service.
"""

import logging
import os
from datetime import datetime

try:
    from onedatafilerestclient import OnedataFileRESTClient
    from onedatafilerestclient.errors import (
        OnedataError,
        OnedataRESTError,
    )
except ImportError:
    OnedataFileRESTClient = None

from galaxy.util import (
    mapped_chars,
    string_as_bool,
)
from ._caching_base import CachingConcreteObjectStore
from .caching import (
    enable_cache_monitor,
    parse_caching_config_dict_from_xml,
)

NO_ONEDATA_ERROR_MESSAGE = (
    "ObjectStore configured to use Onedata, but no OnedataFileRESTClient dependency "
    "available. Please install and properly configure Onedata or modify Object "
    "Store configuration."
)

log = logging.getLogger(__name__)

STREAM_CHUNK_SIZE = 1024 * 1024


def _parse_config_xml(config_xml):
    try:
        auth_xml = _get_config_xml_elements(config_xml, "auth")[0]
        access_token = auth_xml.get("access_token")

        conn_xml = _get_config_xml_elements(config_xml, "connection")[0]
        onezone_domain = conn_xml.get("onezone_domain")
        disable_tls_certificate_validation = string_as_bool(conn_xml.get("disable_tls_certificate_validation", "False"))

        space_xml = _get_config_xml_elements(config_xml, "space")[0]
        space_name = space_xml.get("name")
        galaxy_root_dir = space_xml.get("galaxy_root_dir", "")

        cache_dict = parse_caching_config_dict_from_xml(config_xml)

        extra_dirs = [
            {attr: elem.get(attr) for attr in ("type", "path")}
            for elem in _get_config_xml_elements(config_xml, "extra_dir")
        ]

        return {
            "auth": {"access_token": access_token},
            "connection": {
                "onezone_domain": onezone_domain,
                "disable_tls_certificate_validation": disable_tls_certificate_validation,
            },
            "space": {"name": space_name, "galaxy_root_dir": galaxy_root_dir},
            "cache": cache_dict,
            "extra_dirs": extra_dirs,
            "private": CachingConcreteObjectStore.parse_private_from_config_xml(config_xml),
        }
    except Exception:
        # Toss it back up after logging, we can't continue loading at this point.
        log.exception("Malformed Onedata ObjectStore Configuration XML -- unable to continue")
        raise


def _get_config_xml_elements(config_xml, tag):
    elements = config_xml.findall(tag)

    if not elements:
        msg = f"No {tag} element in config XML tree"
        log.error(msg)
        raise Exception(msg)

    return elements


class OnedataObjectStore(CachingConcreteObjectStore):
    """
    Object store that stores objects as items in an Onedata. A local
    cache exists that is used as an intermediate location for files between
    Galaxy and Onedata.
    """

    store_type = "onedata"

    def __init__(self, config, config_dict):
        super().__init__(config, config_dict)
        self.cache_monitor = None

        auth_dict = config_dict["auth"]
        self.access_token = auth_dict["access_token"]

        connection_dict = config_dict["connection"]
        self.onezone_domain = connection_dict["onezone_domain"]
        self.disable_tls_certificate_validation = connection_dict.get("disable_tls_certificate_validation", False)

        space_dict = config_dict["space"]
        self.space_name = space_dict["name"]
        self.galaxy_root_dir = space_dict.get("galaxy_root_dir", "")

        cache_dict = config_dict.get("cache") or {}
        self.enable_cache_monitor, self.cache_monitor_interval = enable_cache_monitor(config, config_dict)
        self.cache_size = cache_dict.get("size") or self.config.object_store_cache_size
        self.staging_path = cache_dict.get("path") or self.config.object_store_cache_path
        self.cache_updated_data = cache_dict.get("cache_updated_data", True)

        extra_dirs = {e["type"]: e["path"] for e in config_dict.get("extra_dirs", [])}
        self.extra_dirs.update(extra_dirs)

        self._initialize()

    def _initialize(self):
        if OnedataFileRESTClient is None:
            raise Exception(NO_ONEDATA_ERROR_MESSAGE)

        log.debug(
            f"Configuring Onedata connection to {self.onezone_domain} "
            f"(disable_tls_certificate_validation={self.disable_tls_certificate_validation})"
        )

        alt_space_fqn_separators = [mapped_chars["@"]] if "@" in mapped_chars else None
        verify_ssl = not self.disable_tls_certificate_validation
        self._client = OnedataFileRESTClient(
            self.onezone_domain,
            self.access_token,
            alt_space_fqn_separators=alt_space_fqn_separators,
            verify_ssl=verify_ssl,
        )

        self._ensure_staging_path_writable()
        self._start_cache_monitor_if_needed()

    @classmethod
    def parse_xml(cls, config_xml):
        return _parse_config_xml(config_xml)

    def to_dict(self):
        as_dict = super().to_dict()
        as_dict.update(
            {
                "auth": {
                    "access_token": self.access_token,
                },
                "connection": {
                    "onezone_domain": self.onezone_domain,
                    "disable_tls_certificate_validation": self.disable_tls_certificate_validation,
                },
                "space": {"name": self.space_name, "galaxy_root_dir": self.galaxy_root_dir},
                "cache": {
                    "size": self.cache_size,
                    "path": self.staging_path,
                    "cache_updated_data": self.cache_updated_data,
                },
            }
        )
        return as_dict

    def _build_remote_path(self, rel_path):
        return os.path.join(self.galaxy_root_dir, rel_path)

    def _get_remote_size(self, rel_path):
        try:
            remote_path = self._build_remote_path(rel_path)
            return self._client.get_attributes(self.space_name, attributes=["size"], file_path=remote_path)["size"]
        except OnedataError:
            log.exception("Could not get '%s' size from Onedata", rel_path)
            return -1

    def _exists_remotely(self, rel_path):
        try:
            remote_path = self._build_remote_path(rel_path)
            self._client.get_attributes(self.space_name, attributes=["type", "size"], file_path=remote_path)
            return True
        except OnedataError as ex:
            if _is_not_found_onedata_rest_error(ex):
                return False

            log.exception("Trouble checking '%s' existence in Onedata", rel_path)
            return False

    def _download(self, rel_path):
        try:
            dst_path = self._get_cache_path(rel_path)

            log.debug("Pulling file '%s' into cache to %s", rel_path, dst_path)

            remote_path = self._build_remote_path(rel_path)
            file_size = self._client.get_attributes(self.space_name, attributes=["size"], file_path=remote_path)["size"]

            # Test if cache is large enough to hold the new file
            if not self._caching_allowed(rel_path, file_size):
                return False

            with open(dst_path, "wb") as dst:
                for chunk in self._client.iter_file_content(
                    self.space_name, chunk_size=STREAM_CHUNK_SIZE, file_path=remote_path
                ):
                    dst.write(chunk)

            log.debug("Pulled '%s' into cache to %s", rel_path, dst_path)

            return True
        except OnedataError:
            log.exception("Problem downloading file '%s'", rel_path)
            return False

    def _push_to_storage(self, rel_path, source_file=None, from_string=None):
        """
        Push the file pointed to by ``rel_path`` to the object store under ``rel_path``.
        If ``source_file`` is provided, push that file instead while still using
        ``rel_path`` as the path.
        """
        try:
            source_file = source_file or self._get_cache_path(rel_path)
            if os.path.exists(source_file):
                if os.path.getsize(source_file) == 0 and self._exists_remotely(rel_path):
                    log.debug(
                        "Wanted to push file '%s' to Onedata '%s' but its size is 0; skipping.", source_file, rel_path
                    )
                    return True
                else:
                    start_time = datetime.now()
                    log.debug(
                        "Pushing cache file '%s' of size %s bytes under '%s'",
                        source_file,
                        os.path.getsize(source_file),
                        rel_path,
                    )
                    remote_path = self._build_remote_path(rel_path)

                    if not self._exists_remotely(rel_path):
                        file_id = self._client.create_file(
                            self.space_name, file_path=remote_path, file_type="REG", create_parents=True
                        )
                    else:
                        file_id = self._client.get_file_id(self.space_name, file_path=remote_path)

                    if source_file:
                        with open(source_file, "rb") as src:
                            offset = 0
                            while True:
                                chunk = src.read(STREAM_CHUNK_SIZE)
                                if not chunk:
                                    break

                                self._client.put_file_content(
                                    self.space_name, data=chunk, offset=offset, file_id=file_id
                                )
                                offset += len(chunk)
                    else:
                        self._client.put_file_content(
                            self.space_name, data=from_string.encode("utf-8"), file_id=file_id
                        )

                    end_time = datetime.now()
                    log.debug(
                        "Pushed cache file '%s' under '%s' (%s bytes transferred in %s sec)",
                        source_file,
                        rel_path,
                        os.path.getsize(source_file),
                        end_time - start_time,
                    )
                return True
            else:
                log.error("Source file does not exist.", rel_path, source_file)
        except OnedataError:
            log.exception("Trouble pushing Onedata key '%s' from file '%s'", rel_path, source_file)
            raise
        return False

    def _delete_existing_remote(self, rel_path) -> bool:
        try:
            onedata_path = self._build_remote_path(rel_path)
            self._client.remove(self.space_name, file_path=onedata_path)
            return True
        except OnedataError:
            log.exception("Could not delete '%s' from Onedata", rel_path)
            return False

    def _get_object_url(self, _obj, **_kwargs):
        return None

    def _get_store_usage_percent(self, _obj):
        return 0.0

    def shutdown(self):
        self._shutdown_cache_monitor()


def _is_not_found_onedata_rest_error(ex):
    if isinstance(ex, OnedataRESTError):
        if ex.http_code == 404:
            return True

        if ex.http_code == 400 and ex.category == "posix":
            return ex.details["errno"] == "enoent"

    return False
